// Author: Pete Kirkham
//
// quad store utils.
//

#include "utils/string_utils.h"

#include <iostream>

namespace
{
    const char* const hex    ("0123456789abcdef");
    const char* const big_hex("0123456789ABCDEF");
}

std::string to_hex(char value)
{
    char buf[] = {hex[(value>>4)&0xf], hex[value&0xf], 0};
    return buf;
}

std::string to_hex(ssize_t value)
{
    return to_hex(size_t(value));
}

std::string to_hex(void* value)
{
    return to_hex(size_t(value));
}

std::string to_hex(size_t value)
{
    std::ostringstream  buf;

    for (int j(28); j >= 0; j-=4)
    {
        buf << hex[(value >> j)&0xf];
    }

    return buf.str();
}
#ifdef USE_SSE2
std::string to_hex  (const __m128i& value)
{
    int p[4];
    _mm_storeu_si128((__m128i*)(p), value);

    std::ostringstream  buf;

    for (int i(3); i >= 0; --i)
    {
        if (i < 3) buf << '-';
        for (int j(28); j >= 0; j-=4)
        {
            buf << hex[(p[i] >> j)&0xf];
        }
    }

    return buf.str();
}
#endif
std::string to_hex  (const std::string& value)
{
    std::ostringstream  buf;

    for (size_t i(0), l(value.length()); i < l; ++i)
    {
        char ch(value[i]);

        buf << hex[(ch >> 4)&0xf] << hex[ch&0xf];
    }

    return buf.str();
}

std::string to_binary  (const std::string& value)
{
    std::ostringstream  buf;

    for (size_t i(0), l(value.length()); i < l; ++i)
    {
        if (i > 0)
        {
            buf << ':';
        }

        char ch(value[i]);

        buf << ((ch >> 7)&1) << ((ch >> 6)&1) << ((ch >> 5)&1) << ((ch >> 4)&1) << ((ch >> 3)&1) << ((ch >> 2)&1) << ((ch >> 1)&1)  << (ch&1);
    }

    return buf.str();
}

std::string to_binary  (char value)
{
    std::ostringstream  buf;

    buf << ((value >> 7)&1) << ((value >> 6)&1) << ((value >> 5)&1) << ((value >> 4)&1) << ((value >> 3)&1) << ((value >> 2)&1) << ((value >> 1)&1)  << (value&1);

    return buf.str();
}

std::string to_binary  (int value)
{
    return to_binary(size_t(value));
}

std::string to_binary  (size_t value)
{
    std::ostringstream  buf;

    for (int j(31); j >= 0; --j)
    {
        buf << ((value >> j)&1);
    }

    return buf.str();
}

std::string to_trimmed_binary  (int value)
{
    return to_trimmed_binary(size_t(value));
}

std::string to_trimmed_binary  (size_t value)
{
    for (int j(31); j >= 0; --j)
    {
        if ((value >> j)&1)
        {
            std::ostringstream  buf;

            for (; j >= 0; --j)
            {
                buf << ((value >> j)&1);
            }

            return buf.str();
        }
    }

    return "0";
}

#ifdef USE_SSE2
std::string to_i4   (const __m128i& value)
{
    int p[4];
    _mm_storeu_si128((__m128i*)(p), value);

    std::ostringstream  buf;

    buf << "</";

    for (int i(3); i >= 0; --i)
    {
        if (i < 3) buf << ',';
        buf << (p[i]);
    }

    buf << "/>";

    return buf.str();
}
#endif
// encodes an utf-8 string with Turtle escape sequences
std::string turtle_encode (const std::string& value)
{
    std::ostringstream  buf;

    write_turtle_encoded(buf, value);

    return buf.str();
}

void write_turtle_encoded (std::ostream& buf, const std::string& value)
{
    for (size_t i(0), l(value.length()); i < l; ++i)
    {
        char ch(value[i]);

        switch (ch)
        {
            case '\n':  buf << "\\n"; break;
            case '\t':  buf << "\\t"; break;
            case '\r':  buf << "\\r"; break;
            case '\"':  buf << "\\\""; break;
            default:
            {
                if ((ch & 0x80) == 0)
                {
                    buf << ch;
                }
                else // utf-8 escape sequences:
                {
                    int codepoint(ch&0x7f);

                    if ((ch & 0xe0) == 0xc0)        // two byte
                    {
                        codepoint = ((codepoint & 0x1f) << 6) | (value[++i] & 0x3f);
                    }
                    else if ((ch & 0xf0) == 0xe0)   // three byte
                    {
                        codepoint = ((codepoint & 0x0f) << 12) | ((value[++i] & 0x3f) << 6) | (value[++i] & 0x3f);
                    }
                    else if ((ch & 0xf8) == 0xf0)   // four byte
                    {
                        codepoint = ((codepoint & 0x07) << 18) | ((value[++i] & 0x3f) << 12) | ((value[++i] & 0x3f) << 6) | (value[++i] & 0x3f);
                    }
                    else    //  unknown
                    {
                        // do nothing
                    }

                    if (codepoint & 0xff00)
                    {
                        buf << "\\U";
                        buf << big_hex[(codepoint>>20)&0xf] << big_hex[(codepoint>>16)&0xf];
                        buf << big_hex[(codepoint>>12)&0xf] << big_hex[(codepoint>>8)&0xf];
                        buf << big_hex[(codepoint>>4)&0xf] << big_hex[codepoint&0xf];
                    }
                    else
                    {
                        buf << "\\u";
                        buf << big_hex[(codepoint>>12)&0xf] << big_hex[(codepoint>>8)&0xf];
                        buf << big_hex[(codepoint>>4)&0xf] << big_hex[codepoint&0xf];
                    }
                }
            }
        }
    }
}
